# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Exceptions that can be raised by the client.
"""

import re


__all__ = ['Error', 'ConnectionError', 'ConnectTimeout', 'ReadTimeout',
           'RetriesExceeded', 'AuthError', 'ClientAuthError',
           'ServerAuthError', 'ParseError', 'VersionError', 'HTTPError',
           'OperationTimeout', 'StatusTimeout', 'NoUniqueMatch', 'NotFound',
           'MetricsResourceNotFound', 'NotificationError',
           'NotificationJMSError', 'NotificationParseError',
           'NotificationConnectionError', 'NotificationSubscriptionError',
           'SubscriptionNotFound', 'ConsistencyError', 'CeasedExistence',
           'OSConsoleError', 'OSConsoleConnectedError',
           'OSConsoleNotConnectedError', 'OSConsoleWebSocketError',
           'OSConsoleAuthError']


class Error(Exception):
    """
    Abstract base class for exceptions specific to this package.

    Exceptions of this class are not raised; only derived exceptions are
    raised.

    Derived from :exc:`~py:exceptions.Exception`.
    """

    def str_def(self):
        """
        Interface definition for the corresponding method derived exception
        classes.

        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts.

        For the exact format returned by derived exception classes, see the
        same-named methods there.
        """
        raise NotImplementedError


class ConnectionError(Error):
    # pylint: disable=redefined-builtin
    """
    This exception indicates a problem with the connection to the HMC, below
    the HTTP level. HTTP errors are indicated via :exc:`~zhmcclient.HTTPError`.

    A retry by the user code is not likely to be successful, unless connect or
    read retries had been disabled when creating the session (see
    :class:`~zhmcclient.Session`).

    Even though this class has exceptions derived from it, exceptions of this
    class may also be raised (if no other derived class matches the
    circumstances).

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg, details):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          details (Exception):
            The original exception describing details about the error.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        self._details = details

    @property
    def details(self):
        """
        The original exception caught by this package, providing more
        information about the problem.

        This will be one of the following exceptions:

        * Any exception derived from
          :exc:`requests.exceptions.RequestException`.
        """
        return self._details

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return f"{self.__class__.__name__}(message={self.args[0]!r})"

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={};
        """
        return \
            f"classname={self.__class__.__name__!r}; message={self.args[0]!r};"


class ConnectTimeout(ConnectionError):
    """
    This exception indicates that a connection to the HMC timed out after
    exhausting the connect retries (see
    :attr:`zhmcclient.RetryTimeoutConfig.connect_retries`).

    Further retrying by the user code is not likely to be successful, unless
    connect retries had been disabled when creating the session (see
    :class:`~zhmcclient.Session`).

    Derived from :exc:`~zhmcclient.ConnectionError`.
    """

    def __init__(self, msg, details, connect_timeout, connect_retries):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          details (Exception):
            The original exception describing details about the error.

          connect_timeout (:term:`integer`):
            The connect timeout in seconds.

          connect_retries (:term:`integer`):
            The number of connect retries.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg, details)
        self._connect_timeout = connect_timeout
        self._connect_retries = connect_retries

    @property
    def connect_timeout(self):
        """
        :term:`integer`: The connect timeout in seconds.
        """
        return self._connect_timeout

    @property
    def connect_retries(self):
        """
        :term:`integer`: The number of connect retries.
        """
        return self._connect_retries

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"connect_timeout={self.connect_timeout!r}, "
            f"connect_retries={self.connect_retries!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; connect_timeout={}; connect_retries={}; message={};
        """  # noqa: E501
        return (
            f"classname={self.__class__.__name__!r}; "
            f"connect_timeout={self.connect_timeout!r}; "
            f"connect_retries={self.connect_retries!r}; "
            f"message={self.args[0]!r};")


class ReadTimeout(ConnectionError):
    """
    This exception indicates that reading an HTTP response from the HMC timed
    out after exhausting the read retries (see
    :attr:`zhmcclient.RetryTimeoutConfig.read_retries`).

    Further retrying by the user code is not likely to be successful, unless
    read retries had been disabled when creating the session (see
    :class:`~zhmcclient.Session`).

    Derived from :exc:`~zhmcclient.ConnectionError`.
    """

    def __init__(self, msg, details, read_timeout, read_retries):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          details (Exception):
            The original exception describing details about the error.

          read_timeout (:term:`integer`):
            The read timeout in seconds.

          read_retries (:term:`integer`):
            The number of read retries.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg, details)
        self._read_timeout = read_timeout
        self._read_retries = read_retries

    @property
    def read_timeout(self):
        """
        :term:`integer`: The read timeout in seconds.
        """
        return self._read_timeout

    @property
    def read_retries(self):
        """
        :term:`integer`: The number of read retries.
        """
        return self._read_retries

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"read_timeout={self.read_timeout!r}, "
            f"read_retries={self.read_retries!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; read_timeout={}; read_retries={}; message={};
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"read_timeout={self.read_timeout!r}; "
            f"read_retries={self.read_retries!r}; "
            f"message={self.args[0]!r};")


class RetriesExceeded(ConnectionError):
    """
    This exception indicates that the maximum number of retries for connecting
    to the HMC, sending HTTP requests or reading HTTP responses was exceeded,
    for reasons other than connect timeouts (see
    :exc:`~zhmcclient.ConnectTimeout`) or read timeouts (see
    :exc:`~zhmcclient.ReadTimeout`).

    Further retrying by the user code is not likely to be successful, unless
    connect or read retries had been disabled when creating the session (see
    :class:`~zhmcclient.Session`).

    Derived from :exc:`~zhmcclient.ConnectionError`.
    """

    def __init__(self, msg, details, connect_retries):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          details (Exception):
            The original exception describing details about the error.

          connect_retries (:term:`integer`):
            The number of connect retries.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg, details)
        self._connect_retries = connect_retries

    @property
    def connect_retries(self):
        """
        :term:`integer`: The number of connect retries.
        """
        return self._connect_retries

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"connect_retries={self.connect_retries!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; connect_retries={}; message={};
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"connect_retries={self.connect_retries!r}; "
            f"message={self.args[0]!r};")


class AuthError(Error):
    # pylint: disable=abstract-method
    """
    This exception indicates erors related to authentication.

    Exceptions of this class are not raised; only derived exceptions are
    raised.

    Derived from :exc:`~zhmcclient.Error`.
    """
    pass


class ClientAuthError(AuthError):
    """
    This exception indicates an authentication related problem detected on
    the client side.

    Derived from :exc:`~zhmcclient.AuthError`.
    """

    def __init__(self, msg):
        # pylint: disable=useless-super-delegation
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return f"{self.__class__.__name__}(message={self.args[0]!r})"

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={};
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r};")


class ServerAuthError(AuthError):
    """
    This exception indicates an authentication error with the HMC.

    Derived from :exc:`~zhmcclient.AuthError`.
    """

    def __init__(self, msg, details):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          details (Exception):
            The :exc:`~zhmcclient.HTTPError` exception describing the
            error returned by the HMC.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        assert isinstance(details, HTTPError)
        self._details = details

    @property
    def details(self):
        """
        The original exception describing details about the error.

        This may be one of the following exceptions:

        * :exc:`~zhmcclient.HTTPError`
        """
        return self._details

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"details.request_method={self.details.request_method!r}, "
            f"details.request_uri={self.details.request_uri!r}, "
            f"details.http_status={self.details.http_status!r}, "
            f"details.reason={self.details.reason!r})")

    def str_def(self):
        # pylint: disable=line-too-long
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; request_method={}; request_uri={}; http_status={}; reason={}; message={};
        """  # noqa: E501
        # pylint: enable=line-too-long
        return (
            f"classname={self.__class__.__name__!r}; "
            f"request_method={self.details.request_method!r}; "
            f"request_uri={self.details.request_uri!r}; "
            f"http_status={self.details.http_status!r}; "
            f"reason={self.details.reason!r}; message={self.args[0]!r};")


class ParseError(Error):
    """
    This exception indicates a parsing error while processing the JSON payload
    in a response from the HMC.

    Derived from :exc:`~zhmcclient.Error`.

    The error location within the payload is automatically determined by
    parsing the error message for the pattern:

    .. code-block:: text

        : line {line} column {column}
    """

    def __init__(self, msg):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

            This should be the message of the `ValueError` exception raised
            by methods of the :class:`py:json.JSONDecoder` class.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        self._line = None
        self._column = None
        if msg:
            m = re.search(r': line ([0-9]+) column ([0-9]+) ', msg)
            if m:
                self._line = int(m.group(1))
                self._column = int(m.group(2))

    @property
    def line(self):
        """
        :term:`integer`: The 1-based line number of the error location within
        the JSON payload.

        `None` indicates that the error location is not available.
        """
        return self._line

    @property
    def column(self):
        """
        :term:`integer`: The 1-based column number of the error location within
        the JSON payload.

        `None` indicates that the error location is not available.
        """
        return self._column

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"line={self.line!r}, column={self.column!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; line={}; column={}; message={};
        """
        return (
            f"classname={self.__class__.__name__!r}; line={self.line!r}; "
            f"column={self.column!r}; message={self.args[0]!r};")


class VersionError(Error):
    """
    This exception indicates that a function of the client requires a minimum
    HMC API version which is not supported by the HMC.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg, min_api_version, api_version):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          min_api_version (:term:`HMC API version`):
            The minimum HMC API version required to perform the function that
            raised this exception.

          api_version (:term:`HMC API version`):
            The actual HMC API version supported by the HMC.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        self._min_api_version = min_api_version
        self._api_version = api_version

    @property
    def min_api_version(self):
        """
        :term:`HMC API version`: The minimum HMC API version required to
        perform the function that raised this exception.
        """
        return self._min_api_version

    @property
    def api_version(self):
        """
        :term:`HMC API version`: The actual HMC API version supported by the
        HMC.
        """
        return self._api_version

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"min_api_version={self.min_api_version!r}, "
            f"api_version={self.api_version!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; min_api_version={}; api_version={}; message={};
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"min_api_version={self.min_api_version!r}; "
            f"api_version={self.api_version!r}; "
            f"message={self.args[0]!r};")


class HTTPError(Error):
    """
    This exception indicates that the HMC returned an HTTP response with a bad
    HTTP status code.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, body):
        """
        Parameters:

          body (:term:`json object`):
            Body of the HTTP error response.

        ``args[0]`` will be set to the 'message' item of the body, or to `None`
        if not present.
        """
        msg = body.get('message', None)
        super().__init__(msg)
        self._body = body

    @property
    def http_status(self):
        """
        :term:`integer`: Numeric HTTP status code (e.g. 500).

        See :term:`RFC2616` for a list of HTTP status codes and reason phrases.
        """
        return self._body.get('http-status', None)

    @property
    def reason(self):
        """
        :term:`integer`: Numeric HMC reason code.

        HMC reason codes provide more details as to the nature of the error
        than is provided by the HTTP status code. This HMC reason code is
        treated as a sub-code of the HTTP status code and thus must be used in
        conjunction with the HTTP status code to determine the error condition.

        Standard HMC reason codes that apply across the entire API are
        described in section 'Common request validation reason codes' in the
        :term:`HMC API` book. Additional operation-specific reason codes may
        also be documented in the description of specific API operations in the
        :term:`HMC API` book.

        The articial reason code 999 is used when the response from the HMC
        contains an HTML-formatted error message.
        """
        return self._body.get('reason', None)

    @property
    def message(self):
        """
        :term:`string`: Message describing the error.

        This message is not currently localized.
        """
        return self._body.get('message', None)

    @property
    def request_method(self):
        """
        :term:`string`: The HTTP method (DELETE, GET, POST, PUT) that caused
        this error response.
        """
        return self._body.get('request-method', None)

    @property
    def request_uri(self):
        """
        :term:`string`: The URI that caused this error response.
        """
        return self._body.get('request-uri', None)

    @property
    def request_query_parms(self):
        """
        List of query-parm-info objects: URI query parameters specified on the
        request.

        Each query-parm-info object identifies a single query parameter by its
        name and includes its value(s).

        An empty list, if the request did not specify any query parameters.
        """
        return self._body.get('request-query-parms', None)

    @property
    def request_headers(self):
        """
        header-info object: HTTP headers specified on the request.

        An empty list, if the request did not specify any HTTP headers.
        """
        return self._body.get('request-headers', None)

    @property
    def request_authenticated_as(self):
        """
        :term:`string`: Name of the HMC user associated with the API session
        under which the request was issued.

        `None`, if the request was issued without an established session or
        there is no HMC user bound to the session.
        """
        return self._body.get('request-authenticated-as', None)

    @property
    def request_body(self):
        """
        The request body, in the form of a JSON document. Note that, since it
        is in the form of a JSON document, this may not be exactly what was
        submitted by the API client program, but it is semantically equivalent.

        If the request body could not be parsed or some other error prevented
        the creation of a JSON document from the request body, this property
        is `None` and the request body is instead available in the
        :attr:`~zhmcclient.HTTPError.request_body_as_string` property.
        """
        return self._body.get('request-body', None)

    @property
    def request_body_as_string(self):
        """
        :term:`string`: The complete request body, or some portion of the
        request body, exactly as it was submitted by the API client program, if
        the :attr:`~zhmcclient.HTTPError.request_body` property is `None`.
        Otherwise, `None`.

        The :attr:`~zhmcclient.HTTPError.request_body_as_string_partial`
        property indicates whether the complete request body is provided in
        this property.
        """
        return self._body.get('request-body-as-string', None)

    @property
    def request_body_as_string_partial(self):
        """
        :class:`py:bool`: Indicates whether the
        :attr:`~zhmcclient.HTTPError.request_body_as_string` property contains
        only part of the request body (`True`) or the entire request body
        (`False`). `None`, if the
        :attr:`~zhmcclient.HTTPError.request_body_as_string` property is
        `None`.
        """
        return self._body.get('request-body-as-string-partial', None)

    @property
    def stack(self):
        """
        :term:`string`: Internal HMC diagnostic information for the error.

        This field is supplied only on selected 5xx HTTP status codes.
        `None`, if not supplied.
        """
        return self._body.get('stack', None)

    @property
    def error_details(self):
        """
        :term:`string`: A nested object that provides additional
        operation-specific error information. This field is provided by
        selected operations, and the format of the nested object is as
        described by that operation.
        """
        return self._body.get('error-details', None)

    def __str__(self):
        """
        Return a human readable string representation of this exception object.
        """
        stack_txt = f' stack={self.stack!r}' if self.stack else ''
        return (
            f"{self.http_status},{self.reason}: {self.message} "
            f"[{self.request_method} {self.request_uri}]{stack_txt}")

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(http_status={self.http_status!r}, "
            f"reason={self.reason!r}, message={self.message!r}, "
            f"request_method={self.request_method!r}, "
            f"request_uri={self.request_uri!r}, stack={self.stack!r}, ...)")

    def str_def(self):
        # pylint: disable=line-too-long
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; request_method={}; request_uri={}; http_status={}; reason={}; message={};
        """  # noqa: E501
        # pylint: enable=line-too-long
        return (
            f"classname={self.__class__.__name__!r}; "
            f"request_method={self.request_method!r}; "
            f"request_uri={self.request_uri!r}; "
            f"http_status={self.http_status!r}; "
            f"reason={self.reason!r}; message={self.args[0]!r};")


class OperationTimeout(Error):
    """
    This exception indicates that the waiting for completion of an asynchronous
    HMC operation has timed out.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg, operation_timeout):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          operation_timeout (:term:`integer`):
            The operation timeout in seconds.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        self._operation_timeout = operation_timeout

    @property
    def operation_timeout(self):
        """
        :term:`integer`: The operation timeout in seconds.
        """
        return self._operation_timeout

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"operation_timeout={self.operation_timeout!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; operation_timeout={}; message={};
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"operation_timeout={self.operation_timeout!r}; "
            f"message={self.args[0]!r};")


class StatusTimeout(Error):
    """
    This exception indicates that the waiting for reaching a desired LPAR
    or Partition status has timed out.

    The possible status values for an LPAR are:

    * ``"not-activated"`` - The LPAR is not active.
    * ``"not-operating"`` - The LPAR is active but no operating system is
      running in the LPAR.
    * ``"operating"`` - The LPAR is active and an operating system is
      running in the LPAR.
    * ``"exceptions"`` - The LPAR or its CPC has one or more unusual
      conditions.

    The possible status values for a Partition are described in the
    'status' property of the data model for the partition resource in the
    :term:`HMC API` book.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg, actual_status, desired_statuses, status_timeout):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          actual_status (:term:`string`):
            The actual status (at the point in time when the status timeout
            expired).

          desired_statuses (iterable of :term:`string`):
            The desired status values that were supposed to be reached.

          status_timeout (:term:`number`):
            The status timeout (in seconds) that has expired.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        self._actual_status = actual_status
        self._desired_statuses = desired_statuses
        self._status_timeout = status_timeout

    @property
    def actual_status(self):
        """
        :term:`string`: The actual status (at the point in time when the
        status timeout expired).
        """
        return self._actual_status

    @property
    def desired_statuses(self):
        """
        iterable of :term:`string`: The desired status values that were
        supposed to be reached.
        """
        return self._desired_statuses

    @property
    def status_timeout(self):
        """
        :term:`number`: The status timeout (in seconds) that has expired.
        """
        return self._status_timeout

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"actual_status={self.actual_status!r}, "
            f"desired_statuses={self.desired_statuses!r}, "
            f"status_timeout={self.status_timeout!r})")

    def str_def(self):
        # pylint: disable=line-too-long
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; actual_status={}; desired_statuses={}; status_timeout={}; message={};
        """  # noqa: E501
        # pylint: enable=line-too-long
        return (
            f"classname={self.__class__.__name__!r}; "
            f"actual_status={self.actual_status!r}; "
            f"desired_statuses={self.desired_statuses!r}; "
            f"status_timeout={self.status_timeout!r}; "
            f"message={self.args[0]!r};")


class NoUniqueMatch(Error):
    """
    This exception indicates that more than one resource matched the filter
    arguments.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, filter_args, manager, resources):
        """
        Parameters:

          filter_args (dict):
            Dictionary of filter arguments by which the resource was attempted
            to be found. Keys are the resource property names, values are
            the match values for that property.

          manager (:class:`~zhmcclient.BaseManager`):
            The manager of the resource, in whose scope the resource was
            attempted to be found.

            Must not be `None`.

          resources (:term:`iterable` of :class:`~zhmcclient.BaseResource`):
            The resources that did match the filter.

            Must not be `None`.

        ``args[0]`` will be set to an exception message that is automatically
        constructed from the input parameters.
        """
        parent = manager.parent
        if parent:
            in_str = f" in {parent.__class__.__name__} {parent.name!r}"
        else:
            in_str = ""
        resource_uris = [r.uri for r in resources]
        msg = (
            f"Found more than one {manager.resource_class.__name__} using "
            f"filter arguments {filter_args!r}{in_str}, with "
            f"URIs: {resource_uris!r}")
        super().__init__(msg)
        self._filter_args = filter_args
        self._manager = manager
        self._resources = list(resources)
        self._resource_uris = resource_uris

    @property
    def filter_args(self):
        """
        dict: Dictionary of filter arguments by which the resource was
        attempted to be found. Keys are the resource property names, values
        are the match values for that property.
        """
        return self._filter_args

    @property
    def manager(self):
        """
        :class:`~zhmcclient.BaseManager`: The manager of the resource, in whose
        scope the resource was attempted to be found.
        """
        return self._manager

    @property
    def resources(self):
        """
        List of :class:`~zhmcclient.BaseResource`: The resources that matched
        the filter.
        """
        return self._resources

    @property
    def resource_uris(self):
        """
        List of URIs of the resources that matched the filter.
        """
        return self._resource_uris

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        parent = self.manager.parent
        parent_classname = parent.__class__.__name__ if parent else None
        parent_name = parent.name if parent else None
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"resource_classname={self.manager.resource_class.__name__!r}, "
            f"filter_args={self.filter_args!r}, "
            f"parent_classname={parent_classname!r}, "
            f"parent_name={parent_name!r}, "
            f"resource_uris={self.resource_uris!r})")

    def str_def(self):
        # pylint: disable=line-too-long
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; resource_classname={}; filter_args={}; parent_classname={}; manager_name={}; message={}; resource_uris={}
        """  # noqa: E501
        # pylint: enable=line-too-long
        parent = self.manager.parent
        parent_classname = parent.__class__.__name__ if parent else None
        parent_name = parent.name if parent else None
        return (
            f"classname={self.__class__.__name__!r}; "
            f"resource_classname={self.manager.resource_class.__name__!r}; "
            f"filter_args={self.filter_args!r}; "
            f"parent_classname={parent_classname!r}; "
            f"parent_name={parent_name!r}; message={self.args[0]!r}; "
            f"resource_uris={self.resource_uris!r}")


class NotFound(Error):
    """
    This exception indicates that a resource was not found.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, filter_args=None, manager=None, message=None):
        """
        Parameters:

          filter_args (dict):
            Dictionary of filter arguments by which the resource was attempted
            to be found. Keys are the resource property names, values are
            the match values for that property.

            May be `None` if the `message` parameter is `None`.
            Will be ignored if the `message` parameter is not `None`.

          manager (:class:`~zhmcclient.BaseManager`):
            The manager of the resource, in whose scope the resource was
            attempted to be found.

            Must not be `None` if the `message` parameter is `None`.
            Will be ignored if the `message` parameter is not `None`.

          message (string):
            The exception message.

            If `None`, the message will be automatically created from the
            `filter_args` and `manager` parameters.

        ``args[0]`` will be set to an exception message that is automatically
        constructed from the input parameters.
        """
        if message is not None:
            msg = message
            filter_args = None
            manager = None
        else:
            assert manager is not None
            parent = manager.parent
            if parent:
                in_str = f" in {parent.__class__.__name__} {parent.name!r}"
            else:
                in_str = ""
            if filter_args and len(filter_args) == 1 and \
                    manager._name_prop in filter_args:
                msg = (
                    f"Could not find {manager.resource_class.__name__} "
                    f"{filter_args[manager._name_prop]!r}{in_str}.")
            else:
                msg = (
                    f"Could not find {manager.resource_class.__name__} "
                    f"using filter arguments {filter_args!r}{in_str}.")
        super().__init__(msg)
        self._filter_args = filter_args
        self._manager = manager

    @property
    def filter_args(self):
        """
        dict: Dictionary of filter arguments by which the resource was
        attempted to be found. Keys are the resource property names, values
        are the match values for that property.

        Will be `None` if the `message` init parameter was not `None`.
        """
        return self._filter_args

    @property
    def manager(self):
        """
        :class:`~zhmcclient.BaseManager`: The manager of the resource, in whose
        scope the resource was attempted to be found.

        Will be `None` if the `message` init parameter was not `None`.
        """
        return self._manager

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        resource_classname = self.manager.resource_class.__name__ \
            if self.manager else None
        parent = self.manager.parent if self.manager else None
        parent_classname = parent.__class__.__name__ if parent else None
        parent_name = parent.name if parent else None
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"resource_classname={resource_classname!r}, "
            f"filter_args={self.filter_args!r}, "
            f"parent_classname={parent_classname!r}, "
            f"parent_name={parent_name!r})")

    def str_def(self):
        # pylint: disable=line-too-long
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; resource_classname={}; filter_args={}; parent_classname={}; parent_name={}; message={};
        """  # noqa: E501
        # pylint: enable=line-too-long
        resource_classname = self.manager.resource_class.__name__ \
            if self.manager else None
        parent = self.manager.parent if self.manager else None
        parent_classname = parent.__class__.__name__ if parent else None
        parent_name = parent.name if parent else None
        return (
            f"classname={self.__class__.__name__!r}; "
            f"resource_classname={resource_classname!r}; "
            f"filter_args={self.filter_args!r}; "
            f"parent_classname={parent_classname!r}; "
            f"parent_name={parent_name!r}; message={self.args[0]!r};")


class MetricsResourceNotFound(Error):
    # pylint: disable=redefined-builtin
    """
    This exception indicates that the resource referenced by a metric object
    value was not found on the HMC.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg, resource_class, managers):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          resource_class (:class:`~zhmcclient.BaseResource`):
            The zhmcclient resource class of the resource that was not found.

          managers (list of :class:`~zhmcclient.BaseManager`):
            List of zhmcclient resource managers that were searched for the
            resource.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        self._resource_class = resource_class
        self._managers = managers

    @property
    def resource_class(self):
        """
        The zhmcclient resource class of the resource that was not found.
        """
        return self._resource_class

    @property
    def managers(self):
        """
        List of zhmcclient resource managers that were searched for the
        resource
        """
        return self._managers

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return f"{self.__class__.__name__}(message={self.args[0]!r})"

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={}
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r}")


class NotificationError(Error):
    """
    Abstract base class for exceptions raised by
    :class:`~zhmcclient.NotificationListener`.

    Exceptions of this class are not raised; only derived exceptions are
    raised.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def str_def(self):
        """
        Interface definition for the corresponding method derived exception
        classes.

        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts.

        For the exact format returned by derived exception classes, see the
        same-named methods there.
        """
        raise NotImplementedError


class NotificationJMSError(NotificationError):
    """
    This exception indicates that a JMS error was returned by the HMC in
    the notification protocol.

    Derived from :exc:`~zhmcclient.NotificationError`.
    """

    def __init__(self, msg, jms_headers, jms_message):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

            This is either the 'message' item from the JMS headers if present,
            or a generic message.

          jms_headers (dict):
            The JMS headers returned by the HMC.

          jms_message (:term:`string`):
            The JMS message body returned by the HMC.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        self._jms_headers = jms_headers
        self._jms_message = jms_message

    @property
    def jms_headers(self):
        """
        dict: The JMS headers returned by the HMC.
        """
        return self._jms_headers

    @property
    def jms_message(self):
        """
        :term:`string`: The JMS message body returned by the HMC.
        """
        return self._jms_message

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"jms_headers={self.jms_headers!r}, "
            f"jms_message={self.jms_message!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={}
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r};")


class NotificationParseError(NotificationError):
    """
    This exception indicates that the message body of a JMS message could not
    be parsed as JSON format.

    Derived from :exc:`~zhmcclient.NotificationError`.
    """

    def __init__(self, msg, jms_message):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

          jms_message (:term:`string`):
            The JMS message body returned by the HMC.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)
        self._jms_message = jms_message

    @property
    def jms_message(self):
        """
        :term:`string`: The JMS message body returned by the HMC.
        """
        return self._jms_message

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"jms_message={self.jms_message!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={}
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r};")


class NotificationConnectionError(NotificationError):
    """
    This exception indicates an issue with the STOMP connection to the HMC.

    Derived from :exc:`~zhmcclient.NotificationError`.
    """

    def __init__(self, msg):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        # pylint: disable=useless-super-delegation
        super().__init__(msg)

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={}
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r};")


class NotificationSubscriptionError(NotificationError):
    """
    This exception indicates an issue with the STOMP subscription or
    unsubscription to the HMC.

    Derived from :exc:`~zhmcclient.NotificationError`.
    """

    def __init__(self, msg):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        # pylint: disable=useless-super-delegation
        super().__init__(msg)

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={}
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r};")


class SubscriptionNotFound(NotificationError):
    """
    This exception indicates that a subscripton for a topic was not found.

    Derived from :exc:`~zhmcclient.NotificationError`.
    """

    def __init__(self, msg):
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        # pylint: disable=useless-super-delegation
        super().__init__(msg)

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={}
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r};")


class ConsistencyError(Error):
    # pylint: disable=abstract-method
    """
    This exception indicates that an inconsistency has been detected.

    Please report such exceptions in the zhmcclient issue tracker
    at https://github.com/zhmcclient/python-zhmcclient/issues.

    Derived from :exc:`~zhmcclient.Error`.
    """
    pass


class CeasedExistence(Error):
    # pylint: disable=abstract-method
    """
    This exception indicates that the corresponding HMC resource for an
    auto-updated zhmcclient resource no longer exists.

    This exception will only be raised for zhmcclient resources that are
    enabled for :ref:`auto-updating`.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, resource_uri):
        """
        Parameters:

          resource_uri (:term:`string`):
            URI of the resource that no longer exists.

        ``args[0]`` will be set to a default message.
        """
        msg = f"Resource no longer exists: {resource_uri}"
        super().__init__(msg)
        self._resource_uri = resource_uri

    @property
    def resource_uri(self):
        """
        :term:`string`: The URI of the resource that no longer exists.
        """
        return self._resource_uri

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return (
            f"{self.__class__.__name__}(message={self.args[0]!r}, "
            f"resource_uri={self.resource_uri!r})")

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={}; resource_uri={}
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r}; "
            f"resource_uri={self.resource_uri!r}")


class OSConsoleError(Error):
    """
    This exception indicates errors related to OS consoles.

    Exceptions of this class are not raised; only derived exceptions are
    raised.

    Derived from :exc:`~zhmcclient.Error`.
    """

    def __init__(self, msg):
        # pylint: disable=useless-super-delegation
        """
        Parameters:

          msg (:term:`string`):
            A human readable message describing the problem.

        ``args[0]`` will be set to the ``msg`` parameter.
        """
        super().__init__(msg)

    def __repr__(self):
        """
        Return a string with the state of this exception object, for debug
        purposes.
        """
        return f"{self.__class__.__name__}(message={self.args[0]!r})"

    def str_def(self):
        """
        :term:`string`: The exception as a string in a Python definition-style
        format, e.g. for parsing by scripts:

        .. code-block:: text

            classname={}; message={};
        """
        return (
            f"classname={self.__class__.__name__!r}; "
            f"message={self.args[0]!r};")


class OSConsoleConnectedError(OSConsoleError):
    """
    This exception indicates that the WebSocket is connected when it should not
    be connected.
    """
    pass


class OSConsoleNotConnectedError(OSConsoleError):
    """
    This exception indicates that the WebSocket is not connected when it should
    be connected.
    """
    pass


class OSConsoleWebSocketError(OSConsoleError):
    """
    This exception indicates a problem with a WebSocket operation.
    """
    pass


class OSConsoleAuthError(OSConsoleError):
    """
    This exception indicates an authentication issue when logging in to the OS
    console.
    """
    pass



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=protected-access

"""
Unit tests for _manager module.
"""


from datetime import datetime
import time
import re
import warnings
import pytest

from zhmcclient import BaseResource, BaseManager, Session, NotFound, \
    NoUniqueMatch
from zhmcclient._manager import _NameUriCache
from zhmcclient._utils import matches_filters


class MyResource(BaseResource):
    """
    A derived resource for testing purposes.

    It is only needed because BaseManager needs it; it is not subject
    of test in this unit test module.
    """

    # This init method is not part of the external API, so this testcase may
    # need to be updated if the API changes.
    def __init__(self, manager, uri, name=None, properties=None):
        super().__init__(manager, uri, name, properties)


class MyManager(BaseManager):
    """
    A derived resource manager for testing the (abstract) BaseManager class.
    """

    # This init method is not part of the external API, so this testcase may
    # need to be updated if the API changes.
    def __init__(self, session, case_insensitive_names=False):
        super().__init__(
            resource_class=MyResource,
            class_name='myresource',
            session=session,
            parent=None,  # a top-level resource
            base_uri='/api/myresources/',
            oid_prop='fake_object_id',
            uri_prop='fake_uri_prop',
            name_prop='fake_name_prop',
            query_props=[],
            case_insensitive_names=case_insensitive_names)
        self._list_resources = []  # resources to return in list()
        self._list_called = 0  # number of calls to list()

    def list(self, full_properties=False, filter_args=None):
        # This mocked implementation does its work based upon the
        # _list_resources instance variable, and then applies client-side
        # filtering on top of it.
        result_list = []
        for res in self._list_resources:
            if not filter_args or matches_filters(res, filter_args):
                result_list.append(res)
        self._list_called += 1
        return result_list


class TestManager0:
    """
    Tests for the BaseManager class with no initial resources.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = Session(host='fake-host', userid='fake-user',
                               password='fake-pw')

    def test_resource_object_oid(self):
        """
        Test BaseManager.resource_object() with oid input.
        """

        res_mgr = MyManager(self.session)
        res_oid = 'fake-res-id0711'

        # Execute the code to be tested
        res = res_mgr.resource_object(res_oid)

        res_uri = res_mgr._base_uri + '/' + res_oid

        assert isinstance(res, res_mgr.resource_class)
        assert res.uri == res_uri
        assert res.properties[res_mgr._uri_prop] == res_uri
        assert res.properties[res_mgr._oid_prop] == res_oid
        assert res.properties['class'] == res_mgr.class_name
        exp_parent = res_mgr.parent.uri if res_mgr.parent is not None else None
        assert res.properties['parent'] == exp_parent

    def test_resource_object_uri(self):
        """
        Test BaseManager.resource_object() with uri input.
        """

        res_mgr = MyManager(self.session)
        res_oid = 'fake-res-id0711'
        res_uri = '/api/myresources/' + res_oid

        # Execute the code to be tested
        res = res_mgr.resource_object(res_uri)

        assert isinstance(res, res_mgr.resource_class)
        assert res.uri == res_uri
        assert res.properties[res_mgr._uri_prop] == res_uri
        assert res.properties[res_mgr._oid_prop] == res_oid
        assert res.properties['class'] == res_mgr.class_name
        exp_parent = res_mgr.parent.uri if res_mgr.parent is not None else None
        assert res.properties['parent'] == exp_parent

    def test_resource_object_props(self):
        """
        Test BaseManager.resource_object() with additional props.
        """

        res_mgr = MyManager(self.session)
        res_oid = 'fake-res-id0711'
        res_uri = '/api/myresources/' + res_oid

        add_props = {
            'name': 'abc',
            'prop1': 123,
        }

        # Execute the code to be tested
        res = res_mgr.resource_object(res_uri, add_props)

        assert res.properties['name'] == add_props['name']
        assert res.properties['prop1'] == add_props['prop1']


class TestManager1:
    """
    Tests for the BaseManager class with one resource.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = Session(host='fake-host', userid='fake-user',
                               password='fake-pw')
        self.manager = MyManager(self.session)
        self.resource_uri = "/api/fake-uri-1"
        self.resource_name = "fake-name-1"
        self.resource = MyResource(
            self.manager, uri=self.resource_uri,
            properties={
                self.manager._name_prop: self.resource_name,
                "other": "fake-other-1",
            })
        self.manager._list_resources = [self.resource]

    def test_repr(self):
        """Test BaseManager.__repr__()."""
        manager = self.manager

        repr_str = repr(manager)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{manager.__class__.__name__}\s+at\s+'
            rf'0x{id(manager):08x}\s+\(\\n.*',
            repr_str)

    def test_init_properties(self):
        """Test BaseManager properties after initialization."""

        assert self.manager.resource_class == MyResource
        assert self.manager.session == self.session
        assert self.manager.parent is None

    def test_invalidate_cache(self):
        """Test invalidate_cache()."""

        # Populate the cache by finding a resource by name.
        self.manager.find_by_name(self.resource_name)
        assert self.manager._list_called == 1

        # Check that on the second find by name, list() is not called again.
        self.manager.find_by_name(self.resource_name)
        assert self.manager._list_called == 1

        # Invalidate the cache via invalidate_cache().
        self.manager.invalidate_cache()

        # Check that on the third find by name, list() is called again, because
        # the cache had been invalidated.
        self.manager.find_by_name(self.resource_name)
        assert self.manager._list_called == 2

    def test_flush(self):
        """Test flush() and verify that it raises a DeprecationWarning."""

        # Populate the cache by finding a resource by name.
        self.manager.find_by_name(self.resource_name)
        assert self.manager._list_called == 1

        # Check that on the second find by name, list() is not called again.
        self.manager.find_by_name(self.resource_name)
        assert self.manager._list_called == 1

        # Invalidate the cache via flush().
        with warnings.catch_warnings(record=True) as wngs:
            warnings.simplefilter("always")
            self.manager.flush()
        assert len(wngs) == 1
        wng = wngs[0]
        assert issubclass(wng.category, DeprecationWarning), \
            f"Unexpected warnings class: {wng.category}"

        # Check that on the third find by name, list() is called again, because
        # the cache had been invalidated.
        self.manager.find_by_name(self.resource_name)
        assert self.manager._list_called == 2

    def test_list_not_implemented(self):
        """Test that BaseManager.list() raises NotImplementedError."""
        manager = BaseManager(
            resource_class=MyResource,
            class_name='myresource',
            session=self.session,
            parent=None,  # a top-level resource
            base_uri='/api/myresources/',
            oid_prop='fake_object_id',
            uri_prop='fake_uri_prop',
            name_prop='fake_name_prop',
            query_props=[])

        with pytest.raises(NotImplementedError):
            manager.list()


class TestManager2:
    """
    Tests for the BaseManager class with two resources.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = Session(host='fake-host', userid='fake-user',
                               password='fake-pw')
        self.manager = MyManager(self.session)
        self.resource1 = MyResource(
            self.manager,
            uri="/api/fake-uri-1",
            properties={
                self.manager._name_prop: "fake-name-1",
                "other": "fake-other-1",
                "same": "fake-same",
                "int_other": 23,
                "int_same": 42,
            })
        self.resource2 = MyResource(
            self.manager,
            uri="/api/fake-uri-2",
            properties={
                self.manager._name_prop: "fake-name-2",
                "other": "fake-other-2",
                "same": "fake-same",
                "int_other": 24,
                "int_same": 42,
            })
        self.manager._list_resources = [self.resource1, self.resource2]

    def test_findall_name_none(self):
        """Test BaseManager.findall() with no resource matching by the name
        resource property."""
        filter_args = {self.manager._name_prop: "not-exists"}

        resources = self.manager.findall(**filter_args)

        assert len(resources) == 0

    def test_findall_name_one(self):
        """Test BaseManager.findall() with one resource matching by the name
        resource property."""
        filter_args = {self.manager._name_prop: self.resource2.name}

        resources = self.manager.findall(**filter_args)

        assert len(resources) == 1
        assert resources[0].uri == self.resource2.uri
        assert resources[0].name == self.resource2.name

    def test_findall_str_none(self):
        """Test BaseManager.findall() with no resource matching by a
        string-typed (non-name) resource property."""

        resources = self.manager.findall(other="not-exists")

        assert len(resources) == 0

    def test_findall_str_one(self):
        """Test BaseManager.findall() with one resource matching by a
        string-typed (non-name) resource property."""

        resources = self.manager.findall(other="fake-other-2")

        assert len(resources) == 1
        assert resources[0].uri == self.resource2.uri
        assert resources[0].name == self.resource2.name

    def test_findall_str_one_and(self):
        """Test BaseManager.findall() with one resource matching by two
        string-typed (non-name) resource properties (which get ANDed)."""

        resources = self.manager.findall(same="fake-same",
                                         other="fake-other-2")

        assert len(resources) == 1
        assert resources[0].uri == self.resource2.uri
        assert resources[0].name == self.resource2.name

    def test_findall_str_two(self):
        """Test BaseManager.findall() with two resources matching by a
        string-typed (non-name) resource property."""

        resources = self.manager.findall(same="fake-same")

        assert len(resources) == 2
        assert {res.uri for res in resources} == \
            {self.resource1.uri, self.resource2.uri}

    def test_findall_str_two_or(self):
        """Test BaseManager.findall() with two resources matching by a
        list of string-typed (non-name) resource properties (which get
        ORed)."""

        resources = self.manager.findall(other=["fake-other-1",
                                                "fake-other-2"])

        assert len(resources) == 2
        assert {res.uri for res in resources} == \
            {self.resource1.uri, self.resource2.uri}

    def test_findall_int_none(self):
        """Test BaseManager.findall() with no resource matching by a
        integer-typed resource property."""

        resources = self.manager.findall(int_other=815)

        assert len(resources) == 0

    def test_findall_int_one(self):
        """Test BaseManager.findall() with one resource matching by a
        integer-typed resource property."""

        resources = self.manager.findall(int_other=24)

        assert len(resources) == 1
        assert resources[0].uri == self.resource2.uri
        assert resources[0].name == self.resource2.name

    def test_findall_int_two(self):
        """Test BaseManager.findall() with two resources matching by a
        integer-typed resource property."""

        resources = self.manager.findall(int_same=42)

        assert len(resources) == 2
        assert {res.uri for res in resources} == \
            {self.resource1.uri, self.resource2.uri}

    def test_find_name_none(self):
        """Test BaseManager.find() with no resource matching by the name
        resource property."""
        filter_args = {self.manager._name_prop: "not-exists"}

        with pytest.raises(NotFound):
            self.manager.find(**filter_args)

    def test_find_name_one(self):
        """Test BaseManager.find() with one resource matching by the name
        resource property."""
        filter_args = {self.manager._name_prop: self.resource2.name}

        resource = self.manager.find(**filter_args)

        assert resource.uri == self.resource2.uri
        assert resource.name == self.resource2.name

    def test_find_str_none(self):
        """Test BaseManager.find() with no resource matching by a
        string-typed (non-name) resource property."""
        with pytest.raises(NotFound):

            self.manager.find(other="not-exists")

    def test_find_str_one(self):
        """Test BaseManager.find() with one resource matching by a
        string-typed (non-name) resource property."""

        resource = self.manager.find(other="fake-other-2")

        assert resource.uri == self.resource2.uri
        assert resource.name == self.resource2.name

    def test_find_str_two(self):
        """Test BaseManager.find() with two resources matching by a
        string-typed (non-name) resource property."""
        with pytest.raises(NoUniqueMatch):

            self.manager.find(same="fake-same")

    def test_find_int_none(self):
        """Test BaseManager.find() with no resource matching by a
        string-typed (non-name) resource property."""
        with pytest.raises(NotFound):

            self.manager.find(int_other=815)

    def test_find_int_one(self):
        """Test BaseManager.find() with one resource matching by a
        string-typed (non-name) resource property."""

        resource = self.manager.find(int_other=24)

        assert resource.uri == self.resource2.uri
        assert resource.name == self.resource2.name

    def test_find_int_two(self):
        """Test BaseManager.find() with two resources matching by a
        string-typed (non-name) resource property."""
        with pytest.raises(NoUniqueMatch):

            self.manager.find(int_same=42)

    def test_find_by_name_none(self):
        """Test BaseManager.find_by_name() with no resource matching by the
        name resource property."""

        with pytest.raises(NotFound):
            self.manager.find_by_name("not-exists")

    def test_find_by_name_one(self):
        """Test BaseManager.find_by_name() with one resource matching by the
        name resource property."""

        resource = self.manager.find_by_name(self.resource2.name)

        assert resource.uri == self.resource2.uri
        assert resource.name == self.resource2.name


class TestNameUriCache:
    """All tests for the _NameUriCache class."""

    @staticmethod
    def assert_datetime_near(dt1, dt2, max_delta=0.5):
        """
        Assert that the two times are equal within a delta.
        """
        delta = abs(dt2 - dt1).total_seconds()
        assert delta <= max_delta, \
            f"Datetime values are {delta} s apart, maximum is {max_delta} s"

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = Session(host='fake-host', userid='fake-user',
                               password='fake-pw')
        self.manager = MyManager(self.session)
        self.resource1_uri = "/api/fake-uri-1"
        self.resource1_name = "fake-name-1"
        self.resource1 = MyResource(
            self.manager, uri=self.resource1_uri,
            properties={
                self.manager._name_prop: self.resource1_name,
                "other": "fake-other-1"
            })
        self.resource2_uri = "/api/fake-uri-2"
        self.resource2_name = "fake-name-2"
        self.resource2 = MyResource(
            self.manager, uri=self.resource2_uri,
            properties={
                self.manager._name_prop: self.resource2_name,
                "other": "fake-other-2"
            })
        self.all_names = {self.resource1_name, self.resource2_name}
        self.manager._list_resources = [self.resource1, self.resource2]

        self.timetolive = 1.0  # seconds
        self.cache = _NameUriCache(
            self.manager, self.timetolive, case_insensitive_names=False)
        self.created = datetime.now()

    def test_initial(self):
        """Test initial cache state."""

        assert self.cache._manager == self.manager
        assert self.cache._timetolive == self.timetolive
        assert self.cache._uris == {}
        self.assert_datetime_near(self.cache._invalidated, self.created)

    def test_get_no_invalidate(self):
        """Tests for get() without auto-invalidating the cache."""

        # Check that accessing an existing resource name that is not yet in the
        # cache brings all resources into the cache and causes list() to be
        # called once.
        resource1_name, resource1_uri = self.cache.get(self.resource1_name)
        assert resource1_name == self.resource1.name
        assert resource1_uri == self.resource1.uri
        assert set(self.cache._uris.keys()) == self.all_names
        assert self.manager._list_called == 1

        # Check that on the second access of the same name, list() is not
        # called again.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1

    def test_get_non_existing(self):
        """Tests for get() of a non-existing entry."""

        # Check that accessing a non-existing resource name raises an
        # exception, but has populated the cache.
        with pytest.raises(NotFound):
            self.cache.get('non-existing')
        assert set(self.cache._uris.keys()) == self.all_names
        assert self.manager._list_called == 1

    def test_get_auto_invalidate(self):
        """Tests for get() with auto-invalidating the cache."""

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        self.assert_datetime_near(self.cache._invalidated, self.created)

        # Check that on the second access of the same name, list() is not
        # called again.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1

        # Wait until the time-to-live has safely passed.
        time.sleep(self.timetolive + 0.2)

        # Check that on the third access of the same name, list() is called
        # again, because the cache now has auto-invalidated.
        self.cache.get(self.resource1_name)
        invalidated = datetime.now()
        assert self.manager._list_called == 2
        self.assert_datetime_near(self.cache._invalidated, invalidated)

    def test_get_manual_invalidate(self):
        """Tests for get() and manual invalidate()."""

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        self.assert_datetime_near(self.cache._invalidated, self.created)

        # Check that on the second access of the same name, list() is not
        # called again.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1

        # Manually invalidate the cache.
        self.cache.invalidate()
        invalidated = datetime.now()
        self.assert_datetime_near(self.cache._invalidated, invalidated)
        assert self.cache._uris == {}

        # Check that on the third access of the same name, list() is called
        # again, because the cache has been invalidated.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 2
        assert set(self.cache._uris.keys()) == self.all_names

    def test_refresh_empty(self):
        """Test refresh() on an empty cache."""

        # Refresh the cache and check that this invalidates it and
        # re-populates it.
        self.cache.refresh()
        refreshed = datetime.now()
        self.assert_datetime_near(self.cache._invalidated, refreshed)
        assert self.manager._list_called == 1
        assert set(self.cache._uris.keys()) == self.all_names

    def test_refresh_populated(self):
        """Test refresh() on a fully populated cache."""

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        self.assert_datetime_near(self.cache._invalidated, self.created)

        # Refresh the cache and check that this invalidates it and
        # re-populates it.
        self.cache.refresh()
        refreshed = datetime.now()
        self.assert_datetime_near(self.cache._invalidated, refreshed)
        assert self.manager._list_called == 2
        assert set(self.cache._uris.keys()) == self.all_names

    def test_delete_existing(self):
        """Test delete() of an existing cache entry, and re-accessing it."""

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        self.assert_datetime_near(self.cache._invalidated, self.created)

        # Delete an existing cache entry and check that the entry is now gone.
        self.cache.delete(self.resource1_name)
        assert set(self.cache._uris.keys()) == {self.resource2_name}

        # Re-access the deleted entry, and check that list() is called again
        # to get that entry into the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 2
        assert set(self.cache._uris.keys()) == self.all_names

    def test_delete_non_existing(self):
        """Test delete() of a non-existing cache entry."""

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        self.assert_datetime_near(self.cache._invalidated, self.created)

        # Delete a non-existing cache entry and check that no exception is
        # raised and that the cache still contains the same entries.
        self.cache.delete('non-existing')
        assert self.manager._list_called == 1
        assert set(self.cache._uris.keys()) == self.all_names

    def test_delete_none(self):
        """Test delete() of `None`."""

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        self.assert_datetime_near(self.cache._invalidated, self.created)

        # Delete `None` and check that no exception is raised and that the
        # cache still contains the same entries.
        self.cache.delete(None)
        assert self.manager._list_called == 1
        assert set(self.cache._uris.keys()) == self.all_names

    def test_update_from_empty(self):
        """Test update_from() on an empty cache."""

        resource3_uri = "/api/fake-uri-3"
        resource3_name = "fake-name-3"
        resource3 = MyResource(
            self.manager, uri=resource3_uri,
            properties={
                self.manager._name_prop: resource3_name,
            })
        resource4_uri = "/api/fake-uri-4"
        resource4_name = "fake-name-4"
        resource4 = MyResource(
            self.manager, uri=resource4_uri,
            properties={
                self.manager._name_prop: resource4_name,
            })

        # Update the cache from these two resources check that they are now in
        # the cache (and that list() has not been called)
        self.cache.update_from([resource3, resource4])
        assert self.manager._list_called == 0
        assert set(self.cache._uris.keys()) == {resource3_name, resource4_name}

    def test_update_from_populated_modify_name(self):
        """Test update_from() on a populated cache and modify the URI of one
        existing entry."""

        resource3_uri = "/api/fake-uri-3"
        resource3_name = "fake-name-3"
        resource3 = MyResource(
            self.manager, uri=resource3_uri,
            properties={
                self.manager._name_prop: resource3_name,
            })
        resource2_new_uri = "/api/fake-new-uri-2"
        resource2_new = MyResource(
            self.manager, uri=resource2_new_uri,
            properties={
                self.manager._name_prop: self.resource2_name,
            })

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        assert set(self.cache._uris.keys()) == \
            {self.resource1_name, self.resource2_name}

        # Update the cache from these two resources check that they are now in
        # the cache (and that list() has not been called again).
        self.cache.update_from([resource3, resource2_new])
        assert self.manager._list_called == 1
        assert set(self.cache._uris.keys()) == \
            {self.resource1_name, self.resource2_name, resource3_name}

        # Access the modified entry, and check that the entry has changed
        # (and that list() has not been called again).
        resource2_name, resource2_uri = self.cache.get(self.resource2_name)
        assert self.manager._list_called == 1
        assert resource2_uri == resource2_new.uri
        assert resource2_name == resource2_new.name

    def test_update_empty(self):
        """Test update() on an empty cache."""

        resource3_uri = "/api/fake-uri-3"
        resource3_name = "fake-name-3"

        # Update the cache, to get the entry added.
        self.cache.update(resource3_name, resource3_uri)
        assert self.manager._list_called == 0

        # Access the new entry, and check the entry (and that list() has not
        # been called).
        act_resource3_name, act_resource3_uri = self.cache.get(resource3_name)
        assert self.manager._list_called == 0
        assert act_resource3_uri == resource3_uri
        assert act_resource3_name == resource3_name

    def test_update_empty_empty(self):
        """Test update() on an empty cache with an empty resource name."""

        resource3_uri = "/api/fake-uri-3"
        resource3_name = ""

        # Update the cache with the empty resource name, and check that no
        # exception is raised and that the cache is still empty.
        self.cache.update(resource3_name, resource3_uri)
        assert self.cache._uris == {}
        assert self.manager._list_called == 0

    def test_update_empty_none(self):
        """Test update() on an empty cache with a `None` resource name."""

        resource3_uri = "/api/fake-uri-3"
        resource3_name = None

        # Update the cache with the empty resource name, and check that no
        # exception is raised and that the cache is still empty.
        self.cache.update(resource3_name, resource3_uri)
        assert self.cache._uris == {}
        assert self.manager._list_called == 0

    def test_update_populated_new(self):
        """Test update() on a populated cache with a new entry."""

        resource3_uri = "/api/fake-uri-3"
        resource3_name = "fake-name-3"

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        assert set(self.cache._uris.keys()) == \
            {self.resource1_name, self.resource2_name}

        # Update the cache, to get the new entry added.
        self.cache.update(resource3_name, resource3_uri)
        assert self.manager._list_called == 1

        # Access the new entry, and check the entry (and that list() has not
        # been called).
        act_resource3_name, act_resource3_uri = self.cache.get(resource3_name)
        assert self.manager._list_called == 1
        assert act_resource3_uri == resource3_uri
        assert act_resource3_name == resource3_name

    def test_update_populated_modify(self):
        """Test update() on a populated cache by modifying an existing
        entry."""

        resource2_new_uri = "/api/fake-new-uri-2"

        # Populate the cache.
        self.cache.get(self.resource1_name)
        assert self.manager._list_called == 1
        assert set(self.cache._uris.keys()) == \
            {self.resource1_name, self.resource2_name}

        # Update the cache, to get the existing entry modified.
        self.cache.update(self.resource2_name, resource2_new_uri)
        assert self.manager._list_called == 1

        # Access the new entry, and check the entry (and that list() has not
        # been called again).
        act_resource2_name, act_resource2_uri = \
            self.cache.get(self.resource2_name)
        assert self.manager._list_called == 1
        assert act_resource2_uri == resource2_new_uri
        assert act_resource2_name == self.resource2_name


TESTCASES_FINDALL_NAME_PARMS = [
    # Testcases for test_findall_name_parms().
    # Each list item is a tuple defining a testcase in the following format:
    # - res_names: List of names of existing resources
    # - case_insensitive_names(bool): Whether the resource type has case
    #   insensitive name matching.
    # - name_filter: Filter value for matching the resource name
    # - exp_names: Expected resource names in result
    # - exp_exc: Expected exception, or None
    (
        ['name1', 'name2'],
        False,
        'name1',
        ['name1'],
        None
    ),
    (
        ['name1', 'name2'],
        False,
        'name.*',
        ['name1', 'name2'],
        None
    ),
    (
        ['name1', 'name2'],
        False,
        'ame',
        [],
        None
    ),
    (
        ['name-1', 'name.1', 'name1'],
        False,
        'name.1',
        ['name-1', 'name.1'],
        None
    ),
    (
        ['name-1', 'name.1', 'name1'],
        False,
        r'name\.1',
        ['name.1'],
        None
    ),
    (
        ['nax', 'nam', 'name', 'name1'],
        False,
        'name?',
        ['nam', 'name'],
        None
    ),
    (
        ['nax', 'nam', 'name', 'namee', 'name1'],
        False,
        'name*',
        ['nam', 'name', 'namee'],
        None
    ),
    (
        ['nax', 'nam', 'name', 'namee', 'name1'],
        False,
        'name+',
        ['name', 'namee'],
        None
    ),
    (
        ['nax', 'naM', 'Name', 'nAmee', 'Name1'],
        True,
        'Nam',
        ['naM'],
        None
    ),
    (
        ['nax', 'naM', 'Name', 'nAmee', 'Name1'],
        True,
        'Name+',
        ['Name', 'nAmee'],
        None
    ),
]


@pytest.mark.parametrize(
    "res_names, case_insensitive_names, name_filter, exp_names, exp_exc",
    TESTCASES_FINDALL_NAME_PARMS)
def test_findall_name_parms(
        res_names, case_insensitive_names, name_filter, exp_names, exp_exc):
    """
    Test BaseManager.findall() with filter args for matching the resource name.
    """
    session = Session(host='fake-host', userid='fake-user', password='fake-pw')
    manager = MyManager(session, case_insensitive_names=case_insensitive_names)
    all_resources = []
    for res_name in res_names:
        uri = res_name + '-uri'
        res = MyResource(
            manager,
            uri=uri,
            properties={
                manager._name_prop: res_name,
                "other": "fake-other",
                "same": "fake-same",
                "int_other": 23,
                "int_same": 42,
            })
        all_resources.append(res)
    # pylint: disable=attribute-defined-outside-init
    manager._list_resources = all_resources

    filter_args = {manager._name_prop: name_filter}

    if exp_exc:
        with pytest.raises(exp_exc):

            manager.findall(**filter_args)

    else:

        resources = manager.findall(**filter_args)

        act_names = [_res.name for _res in resources]
        assert act_names == exp_names


TESTCASES_FIND_NAME_PARMS = [
    # Testcases for test_find_name_parms().
    # Each list item is a tuple defining a testcase in the following format:
    # - res_names: List of names of existing resources
    # - case_insensitive_names(bool): Whether the resource type has case
    #   insensitive name matching.
    # - name_filter: Filter value for matching the resource name
    # - exp_name: Expected resource name in result
    # - exp_exc: Expected exception, or None
    (
        ['name1', 'name2'],
        False,
        'name1',
        'name1',
        None
    ),
    (
        ['name1', 'name2'],
        False,
        'name.*',
        None,
        NotFound
    ),
    (
        ['name1', 'name2'],
        False,
        'ame',
        None,
        NotFound
    ),
    (
        ['name-1', 'name.1', 'name1'],
        False,
        'name.1',
        'name.1',
        None
    ),
    (
        ['name-1', 'name.1', 'name1'],
        False,
        r'name\.1',
        None,
        NotFound
    ),
    (
        ['nAme1', 'name2'],
        True,
        'Name1',
        'nAme1',
        None
    ),
]


@pytest.mark.parametrize(
    "res_names, case_insensitive_names, name_filter, exp_name, exp_exc",
    TESTCASES_FIND_NAME_PARMS)
def test_find_name_parms(
        res_names, case_insensitive_names, name_filter, exp_name, exp_exc):
    """
    Test BaseManager.find() with filter args for matching the resource name.
    """

    session = Session(host='fake-host', userid='fake-user', password='fake-pw')
    manager = MyManager(session, case_insensitive_names=case_insensitive_names)
    all_resources = []
    for res_name in res_names:
        uri = res_name + '-uri'
        res = MyResource(
            manager,
            uri=uri,
            properties={
                manager._name_prop: res_name,
                "other": "fake-other",
                "same": "fake-same",
                "int_other": 23,
                "int_same": 42,
            })
        all_resources.append(res)
    # pylint: disable=attribute-defined-outside-init
    manager._list_resources = all_resources

    filter_args = {manager._name_prop: name_filter}

    if exp_exc:
        with pytest.raises(exp_exc):

            manager.find(**filter_args)

    else:

        resource = manager.find(**filter_args)

        act_name = resource.name
        assert act_name == exp_name



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for users (on CPCs in DPM mode).

These tests do not change any existing users, but create,
modify and delete test users.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic User objects (e.g. find_by_name())
USER_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in User objects returned by list() without full props
USER_LIST_PROPS = ['object-uri', 'name', 'type']

# Properties whose values can change between retrievals of User objects
USER_VOLATILE_PROPS = []


def test_user_find_list(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "users")

    # Pick the users to test with
    user_list = console.users.list()
    if not user_list:
        skip_warn(f"No users defined on HMC {hd.host}")
    user_list = pick_test_resources(user_list)

    for user in user_list:
        print(f"Testing with user {user.name!r}")
        runtest_find_list(
            hmc_session, console.users, user.name, 'name',
            'object-uri', USER_VOLATILE_PROPS, USER_MINIMAL_PROPS,
            USER_LIST_PROPS)


def test_user_property(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "users")

    # Pick the users to test with
    user_list = console.users.list()
    if not user_list:
        skip_warn(f"No users defined on HMC {hd.host}")
    user_list = pick_test_resources(user_list)

    for user in user_list:
        print(f"Testing with user {user.name!r}")

        # Select a property that is not returned by list()
        non_list_prop = 'description'

        runtest_get_properties(user.manager, non_list_prop)


def test_user_crud(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a user.
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "users")

    user_name = TEST_PREFIX + '_test_user_crud_user1'
    user_name_new = user_name + '_new'
    pwrule_name = TEST_PREFIX + '_test_user_crud_pwrule1'

    # Ensure a clean starting point for this test
    try:
        user = console.users.find(name=user_name)
    except zhmcclient.NotFound:
        pass
    else:
        warnings.warn(
            f"Deleting test user from previous run: {user_name!r}",
            UserWarning)
        user.delete()
    try:
        pwrule = console.password_rules.find(name=pwrule_name)
    except zhmcclient.NotFound:
        pass
    else:
        warnings.warn(
            f"Deleting test password rule from previous run: {pwrule_name!r}",
            UserWarning)
        pwrule.delete()

    # Pick a password rule for the user
    try:
        pwrule = console.password_rules.find(name='Basic')
    except zhmcclient.NotFound:
        skip_warn("Password rule 'Basic' not found to create test user")

    # Test creating the user

    user_input_props = {
        'name': user_name,
        'description': 'Test user for zhmcclient end2end tests',
        'type': 'standard',
        'authentication-type': 'local',
        'password-rule-uri': pwrule.uri,
        'password': 'Abcd',  # Basic rule: 4-8 alphanumeric
    }
    user_auto_props = {
        'disabled': False,
    }
    task_name = 'Manage Users' \
        if user_input_props['type'] == 'standard' \
        else 'Manage User Templates'

    # The code to be tested
    try:
        user = console.users.create(user_input_props)
    except zhmcclient.HTTPError as exc:
        if exc.http_status == 403 and exc.reason == 1:
            skip_warn(
                f"HMC userid {hd.userid!r} is not authorized for task "
                f"{task_name!r} on HMC {hd.host}")
        else:
            raise

    for pn, exp_value in user_input_props.items():
        if pn == 'password':
            continue
        assert user.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"
    user.pull_full_properties()
    for pn, exp_value in user_input_props.items():
        if pn == 'password':
            continue
        assert user.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"
    for pn, exp_value in user_auto_props.items():
        assert user.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"

    # Test updating a property of the user

    new_desc = "Updated user description."

    # The code to be tested
    user.update_properties(dict(description=new_desc))

    assert user.properties['description'] == new_desc
    user.pull_full_properties()
    assert user.properties['description'] == new_desc

    # Test that users cannot be renamed

    with pytest.raises(zhmcclient.HTTPError) as exc_info:

        # The code to be tested
        user.update_properties(dict(name=user_name_new))

    exc = exc_info.value
    assert exc.http_status == 400
    assert exc.reason == 6
    with pytest.raises(zhmcclient.NotFound):
        console.users.find(name=user_name_new)

    # Test deleting the user

    # The code to be tested
    user.delete()

    with pytest.raises(zhmcclient.NotFound):
        console.users.find(name=user_name)



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An :term:`Adapter` is a physical adapter card (e.g. OSA-Express adapter,
Crypto adapter) or a logical adapter (e.g. HiperSockets switch).

Adapter resources are contained in :term:`CPC` resources.

Adapters only exist in CPCs that are in DPM mode.

There are four types of Adapters:

1. Network Adapters:
   Network adapters enable communication through different networking
   transport protocols. These network adapters are OSA-Express,
   HiperSockets and RoCE-Express.
   DPM automatically discovers OSA-Express and RoCE-Express adapters
   because they are physical cards that are installed on the CPC.
   In contrast, HiperSockets are logical adapters and must be
   created and configured by an administrator using the 'Create Hipersocket'
   operation (see create_hipersocket()).
   Network Interface Cards (NICs) provide a partition with access to networks.
   Each NIC represents a unique connection between the partition
   and a specific network adapter.

2. Storage Adapters:
   Fibre Channel connections provide high-speed connections between CPCs
   and storage devices.
   DPM automatically discovers any storage adapters installed on the CPC.
   Host bus adapters (HBAs) provide a partition with access to external
   storage area networks (SANs) and devices that are connected to a CPC.
   Each HBA represents a unique connection between the partition
   and a specific storage adapter.

3. Accelerator Adapters:
   Accelerator adapters provide specialized functions to
   improve performance or use of computer resource like the IBM System z
   Enterprise Data Compression (zEDC) feature.
   DPM automatically discovers accelerators that are installed on the CPC.
   An accelerator virtual function provides a partition with access
   to zEDC features that are installed on a CPC.
   Each virtual function represents a unique connection between
   the partition and a physical feature card.

4. Crypto Adapters:
   Crypto adapters provide cryptographic processing functions.
   DPM automatically discovers cryptographic features that are installed
   on the CPC.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._port import PortManager
from ._logging import logged_api_call
from ._utils import repr_dict, repr_manager, repr_timestamp, matches_filters, \
    divide_filter_args, make_query_str, RC_ADAPTER, repr_obj_id

__all__ = ['AdapterManager', 'Adapter']


class AdapterManager(BaseManager):
    """
    Manager providing access to the :term:`Adapters <Adapter>` in a particular
    :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Cpc` object (in DPM mode):

    * :attr:`~zhmcclient.Cpc.adapters`

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, cpc):
        # This function should not go into the docs.
        # Parameters:
        #   cpc (:class:`~zhmcclient.Cpc`):
        #     CPC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',

            # The adapter-id property is supported for filtering, but due to
            # a firmware defect, adapters with a hex digit in their adapter-id
            # property are not found. Disabling the property causes it to
            # be handled via client-side filtering, so that mitigates the
            # defect.
            # TODO: Re-enable the property once the defect is fixed and the fix
            # is rolled out broadly enough.
            # 'adapter-id',

            'adapter-family',
            'type',
            'status',
        ]

        super().__init__(
            resource_class=Adapter,
            class_name=RC_ADAPTER,
            session=cpc.manager.session,
            parent=cpc,
            base_uri='/api/adapters',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: :term:`CPC` defining the scope for this
        manager.
        """
        return self._parent

    @logged_api_call
    # pylint: disable=arguments-differ
    def list(self, full_properties=False, filter_args=None,
             additional_properties=None):
        """
        List the Adapters in this CPC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to any Adapter to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            This parameter requires HMC 2.16.0 or higher.

        Returns:

          : A list of :class:`~zhmcclient.Adapter` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'adapters'
        list_uri = f'{self.cpc.uri}/adapters'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args,
            additional_properties)

    @logged_api_call
    def create_hipersocket(self, properties):
        """
        Create and configure a HiperSockets Adapter in this CPC.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the scoping CPC.
        * Task permission to the "Create HiperSockets Adapter" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Hipersocket' in the :term:`HMC API` book.

        Returns:

          :class:`~zhmcclient.Adapter`:
            The resource object for the new HiperSockets Adapter.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.cpc.uri + '/adapters', body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        adapter = Adapter(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return adapter


class Adapter(BaseResource):
    """
    Representation of an :term:`Adapter`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of an Adapter, see section 'Data model' in section
    'Adapter object' in the :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.AdapterManager`).

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    # Name of property for port URIs, dependent on adapter family
    port_uris_prop_by_family = {
        'ficon': 'storage-port-uris',
        'osa': 'network-port-uris',
        'roce': 'network-port-uris',
        'hipersockets': 'network-port-uris',
        'cna': 'network-port-uris',
        'cloud-network': 'network-port-uris',  # for preliminary driver
    }

    # URI segment for port URIs, dependent on adapter family
    port_uri_segment_by_family = {
        'ficon': 'storage-ports',
        'osa': 'network-ports',
        'roce': 'network-ports',
        'hipersockets': 'network-ports',
        'cna': 'network-ports',
        'cloud-network': 'network-ports',  # for preliminary driver
    }

    # Port type, dependent on adapter family
    port_type_by_family = {
        'ficon': 'storage',
        'osa': 'network',
        'roce': 'network',
        'hipersockets': 'network',
        'cna': 'network',
        'cloud-network': 'network',  # for preliminary driver
    }

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.AdapterManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, AdapterManager), (
            f"Adapter init: Expected manager type {AdapterManager}, got "
            f"{type(manager)}")
        super().__init__(manager, uri, name, properties)
        # The manager objects for child resources (with lazy initialization):
        self._ports = None
        self._port_uris_prop = None
        self._port_uri_segment = None

    @property
    def ports(self):
        """
        :class:`~zhmcclient.PortManager`: Access to the :term:`Ports <Port>` of
        this Adapter.
        """
        # We do here some lazy loading.
        if not self._ports:
            family = self.get_property('adapter-family')
            try:
                port_type = self.port_type_by_family[family]
            except KeyError:
                port_type = None
            self._ports = PortManager(self, port_type)
        return self._ports

    @property
    def port_uris_prop(self):
        """
        :term:`string`: Name of adapter property that specifies the adapter
        port URIs, or the empty string ('') for adapters without ports.

        For example, 'network-port-uris' for a network adapter.
        """
        if self._port_uris_prop is None:
            family = self.get_property('adapter-family')
            try:
                self._port_uris_prop = self.port_uris_prop_by_family[family]
            except KeyError:
                self._port_uris_prop = ''
        return self._port_uris_prop

    @property
    def port_uri_segment(self):
        """
        :term:`string`: Adapter type specific URI segment for adapter port
        URIs, or the empty string ('') for adapters without ports.

        For example, 'network-ports' for a network adapter.
        """
        if self._port_uri_segment is None:
            family = self.get_property('adapter-family')
            try:
                self._port_uri_segment = self.port_uri_segment_by_family[
                    family]
            except KeyError:
                self._port_uri_segment = ''
        return self._port_uri_segment

    @property
    @logged_api_call
    def maximum_crypto_domains(self):
        """
        Integer: The maximum number of crypto domains on this crypto adapter.

        The following table shows the maximum number of crypto domains for
        crypto adapters supported on IBM Z machine generations in DPM mode. The
        corresponding LinuxONE machine generations are listed in the notes
        below the table:

        =================  ================================  ===============
        Adapter type       Machine generations               Maximum domains
        =================  ================================  ===============
        Crypto Express 5S  z15 (5) / z14 (3) / z13 (1)             85
        Crypto Express 5S  z15 (6) / z14-ZR1 (4) / z13s (2)        40
        Crypto Express 6S  z16 (7) / z15 (5) / z14 (3)             85
        Crypto Express 6S  z15 (6) / z14-ZR1 (4)                   40
        Crypto Express 7S  z16 (7) / z15 (5)                       85
        Crypto Express 7S  z15 (6)                                 40
        Crypto Express 8S  z16 (7)                                 85
        =================  ================================  ===============

        Notes:

        (1) Supported for z13 and LinuxONE Emperor
        (2) Supported for z13s and LinuxONE Rockhopper
        (3) Supported for z14 and LinuxONE Emperor II
        (4) Supported for z14-ZR1 and LinuxONE Rockhopper II
        (5) Supported for z15-T01 and LinuxONE III LT1
        (6) Supported for z15-T02 and LinuxONE III LT2
        (7) Supported for z16-A01 and LinuxONE 4

        If this adapter is not a crypto adapter, `None` is returned.

        If the crypto adapter card type is not known, :exc:`ValueError` is
        raised.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`ValueError`: Unknown crypto card type
        """
        if self.get_property('adapter-family') != 'crypto':
            return None
        card_type = self.get_property('detected-card-type')
        if card_type.startswith('crypto-express-'):
            max_domains = self.manager.cpc.maximum_active_partitions
        else:
            raise ValueError(f"Unknown crypto card type: {card_type!r}")
        return max_domains

    @logged_api_call
    def delete(self):
        """
        Delete this Adapter.

        The Adapter must be a HiperSockets Adapter.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the HiperSockets Adapter to be deleted.
        * Task permission to the "Delete HiperSockets Adapter" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Adapter.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the Adapter.
        * Task permission for the "Adapter Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'Adapter object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    def __repr__(self):
        """
        Return a string with the state of this Adapter, for debug purposes.
        """
        ret = (
            f"{repr_obj_id(self)} (\n"
            f"  _manager={repr_obj_id(self._manager)},\n"
            f"  _uri={self._uri!r},\n"
            f"  _full_properties={self._full_properties!r},\n"
            "  _properties_timestamp="
            f"{repr_timestamp(self._properties_timestamp)},\n"
            f"  _properties={repr_dict(self._properties, indent=4)},\n"
            f"  _ports(lazy)={repr_manager(self._ports, indent=2)}\n"
            ")")
        return ret

    @logged_api_call
    def change_crypto_type(self, crypto_type, zeroize=None):
        """
        Reconfigures a cryptographic adapter to a different crypto type.
        This operation is only supported for cryptographic adapters.

        The cryptographic adapter must be varied offline before its crypto
        type can be reconfigured.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Adapter.
        * Task permission to the "Adapter Details" task.

        Parameters:

          crypto_type (:term:`string`):
            - ``"accelerator"``: Crypto Express5S Accelerator
            - ``"cca-coprocessor"``: Crypto Express5S CCA Coprocessor
            - ``"ep11-coprocessor"``: Crypto Express5S EP11 Coprocessor

          zeroize (bool):
            Specifies whether the cryptographic adapter will be zeroized when
            it is reconfigured to a crypto type of ``"accelerator"``.
            `None` means that the HMC-implemented default of `True` will be
            used.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'crypto-type': crypto_type}
        if zeroize is not None:
            body['zeroize'] = zeroize
        self.manager.session.post(
            self.uri + '/operations/change-crypto-type', resource=self,
            body=body)

    @logged_api_call
    def change_adapter_type(self, adapter_type):
        """
        Reconfigures an adapter from one type to another, or to ungonfigured.
        Currently, only storage adapters can be reconfigured, and their adapter
        type is the supported storage protocol (FCP vs. FICON).

        Storage adapter instances (i.e. :class:`~zhmcclient.Adapter` objects)
        represent daughter cards on a physical storage card. Current storage
        cards require both daughter cards to be configured to the same
        protocol, so changing the type of the targeted adapter will also change
        the type of the adapter instance that represents the other daughter
        card on the same physical card. Zhmcclient users that need to determine
        the related adapter instance can do so by finding the storage adapter
        with a matching first 9 characters (card ID and slot ID) of their
        `card-location` property values.

        The targeted adapter and its related adapter on the same storage card
        must not already have the desired adapter type, they must not be
        attached to any partition, and they must not have an adapter status
        of 'exceptions'.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Adapter.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          adapter_type (:term:`string`):
            - ``"fcp"``: FCP (Fibre Channel Protocol)
            - ``"fc"``: FICON (Fibre Connection) protocol
            - ``"not-configured"``: No adapter type configured

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'type': adapter_type}
        self.manager.session.post(
            self.uri + '/operations/change-adapter-type', resource=self,
            body=body)

    def dump(self):
        """
        Dump this Adapter resource with its properties and child resources
        (recursively) as a resource definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},

                # Child resources:
                "ports": [...],
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        # Dump the child resources
        ports = self.ports.dump()
        if ports:
            resource_dict['ports'] = ports

        return resource_dict

    @logged_api_call
    def list_assigned_partitions(self, full_properties=False, filter_args=None):
        """
        List the partitions assigned to this adapter.

        This method is not supported for OSA adapters configured as OSM
        (because those cannot be assigned to partitions).

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Adapter.

        Parameters:

          full_properties (bool):
            Controls whether the full set of partition properties should be
            retrieved, vs. only a short set (uri, name, status).

          filter_args (dict):
            Filter arguments that narrow the list of returned partitions to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all assigned partitions
            are returned.

        Returns:

          : A list of :class:`~zhmcclient.Partition` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_props = ['name', 'status']
        query_parms, client_filters = divide_filter_args(
            query_props, filter_args)
        query_parms_str = make_query_str(query_parms)
        uri = (f'{self.uri}/operations/get-partitions-assigned-to-adapter'
               f'{query_parms_str}')

        result = self.manager.session.get(uri, resource=self)

        partition_mgr = self.manager.parent.partitions
        resource_obj_list = []
        for props in result['partitions-assigned-to-adapter']:

            # pylint: disable=protected-access
            resource_obj = partition_mgr.resource_class(
                manager=partition_mgr,
                uri=props[partition_mgr._uri_prop],
                name=props.get(partition_mgr._name_prop, None),
                properties=props)

            if matches_filters(resource_obj, client_filters):
                resource_obj_list.append(resource_obj)
                if full_properties:
                    resource_obj.pull_full_properties()

        return resource_obj_list

    @logged_api_call
    def list_sibling_adapters(self, full_properties=False):
        """
        List the other Adapters on the same adapter card as this Adapter.

        Some adapter cards are represented as multiple Adapter objects
        (for example, 2-port FICON Express cards, or 2-port CNA cards).
        This method lists the other Adapter objects that are on the same
        adapter card as this Adapter object.

        This is useful for example to determine the affected Adapter objects
        when replacing the adapter card, or when changing the type of a FICON
        Express adepter (see :meth:`~zhmcclient.Adapter.change_adapter_type`).

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to any Adapter to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

        Returns:

          : A list of :class:`~zhmcclient.Adapter` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # This algorithm is based on the fact that physical adapter cards
        # have their PCHIDs always within a range of 4 adjacent PCHIDs that
        # start at a multiple of 4.

        self_pchid = int(self.prop('adapter-id'), 16)
        if self_pchid >= int('7c0', 16):
            # A virtual adapter with a single PCHID -> no siblings
            return []

        # A physical adapter with a total of 4 PCHIDs reserved for the slot
        pchid_base = self_pchid // 4 * 4
        sibling_pchids = list(range(pchid_base, pchid_base + 4))
        sibling_pchids.remove(self_pchid)
        sibling_adapter_ids = [f'{p:03x}' for p in sibling_pchids]
        filter_args = {'adapter-id': sibling_adapter_ids}
        sibling_adapters = self.manager.cpc.adapters.list(
            full_properties, filter_args)
        return sibling_adapters



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Port` is a physical connector port (jack) of an :term:`Adapter`.

Port resources are contained in Adapter resources.

Ports only exist in :term:`CPCs <CPC>` that are in DPM mode.
"""


import copy
from collections import OrderedDict

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_NETWORK_PORT, RC_STORAGE_PORT

__all__ = ['PortManager', 'Port']

# Resource class names, by port type:
PORT_CLASSES = {
    'network': RC_NETWORK_PORT,
    'storage': RC_STORAGE_PORT,
    None: '',
}


class PortManager(BaseManager):
    """
    Manager providing access to the :term:`Ports <Port>` of a particular
    :term:`Adapter`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible as properties in higher level resources (in this case, the
    :class:`~zhmcclient.Adapter` object).

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, adapter, port_type):
        # This function should not go into the docs.
        # Parameters:
        #   adapter (:class:`~zhmcclient.Adapter`):
        #     Adapter defining the scope for this manager.
        #   port_type (string):
        #     Type of Ports managed by this manager:
        #     * `network`: Ports of a network adapter
        #     * `storage`: Ports of a storage adapter
        #     * None: Adapter family without ports

        try:
            port_class = PORT_CLASSES[port_type]
        except KeyError:
            raise ValueError(f"Unknown port type: {port_type}")

        super().__init__(
            resource_class=Port,
            session=adapter.manager.session,
            class_name=port_class,
            parent=adapter,
            base_uri='',
            # TODO: Re-enable the following when unit/test_hba.py has been
            # converted to using the zhmcclient mock support:
            # base_uri=f'{adapter.uri}/{adapter.port_uri_segment}'
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=[],
            list_has_name=False)

        self._port_type = port_type

    @property
    def adapter(self):
        """
        :class:`~zhmcclient.Adapter`: :term:`Adapter` defining the scope for
        this manager.
        """
        return self._parent

    @property
    def port_type(self):
        """
        :term:`string`: Type of the Ports managed by this object:

        * ``'network'`` - Ports of a network adapter
        * ``'storage'`` - Ports of a storage adapter
        * ``None`` - Adapter family without ports
        """
        return self._port_type

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the Ports of this Adapter.

        If the adapter does not have any ports, an empty list is returned.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the corresponding array property for this resource in the
          parent object is used to list the resources, and the provided filter
          arguments are applied.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Adapter.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.Port` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        uris_prop = self.adapter.port_uris_prop
        if not uris_prop:
            # Adapter does not have any ports
            return []

        return self._list_with_parent_array(
            self.adapter, uris_prop, full_properties, filter_args)


class Port(BaseResource):
    """
    Representation of a :term:`Port`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of a Port, see section
    'Data model - Port Element Object' in section 'Adapter object' in the
    :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.PortManager`).

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.PortManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, PortManager), (
            f"Port init: Expected manager type {PortManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Port.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to the Adapter of this Port.
        * Task permission to the "Adapter Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model - Port Element Object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        # Attempts to change the 'name' property will be rejected by the HMC,
        # so we don't need to update the name-to-URI cache.
        assert self.manager._name_prop not in properties
        self.update_properties_local(copy.deepcopy(properties))

    def dump(self):
        """
        Dump the Port resource with its properties as a resource definition.

        If the adapter of this port is a FICON adapter in the not-configured
        state, the port properties cannot be retrieved from the HMC, so
        an empty dict is returned.

        Otherwise, the returned resource definition has the following format::

            {
                "properties": {...},
            }

        Returns:

          dict: Resource definition of this Port resource.
        """
        resource_dict = OrderedDict()
        adapter = self.manager.parent

        if adapter.prop('type') != 'not-configured':
            # "Get Storage Port Properties" on the port of an unconfigured FICON
            # adapter would return HTTP 404,4: "Get for Storage Port Properties
            # is not supported for this card type".
            self.pull_full_properties()
            resource_dict['properties'] = OrderedDict(self._properties)
            # No child resources
        return resource_dict
